
/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: test.cpp,v 1.8 2004/01/13 15:59:47 whuang Exp $
*/

//
// Test program for dolt
// Creates a variety of drawableObjects and displays them,
//  and demonstrates keyboard handling and multiple scenes
//

#include <dolt-lite.h>
#include <math.h>		// for M_PI

// Windoze math.h doesn't define M_PI
#ifndef M_PI
#define M_PI 3.1415926535897932384626433833
#endif

using namespace dolt;
using namespace std;

Graphics g, g2;

// keyboard handler for primary scene
void gkeyboard( unsigned char key )
{
  switch( key ) {
  case 'q':
    exit( 0 );  // quit
  case 'h':
    cout << "Commands:" << endl
      << " q: Quit" << endl
      << " h: Help" << endl
      << " s: Switch Scenes" << endl
      << " p: Dump to test.ps" << endl;
    g.printControlKeyBindings();
    break;
  case 's':
    Graphics::setCurrentScene( g2 );  // switch scenes
    break;
  case 'p':
    {
      psFile f("test.ps");  // open postscript file
      g.writePS( f );       // write postscript file
    }
    break;
  }
}

// keyboard handler for secondary scene
void g2keyboard( unsigned char key )
{
  switch( key ) {
  case 'q':
    exit( 0 );
  case 'h':
    cout << "Commands:" << endl
      << " q: Quit" << endl
      << " h: Help" << endl
      << " s: Switch Scenes" << endl
      << " p: Dump to test.ps" << endl;
    g2.printControlKeyBindings();
    break;
  case 's':
    Graphics::setCurrentScene( g );
    break;
  case 'p':
    {
      psFile f("test.ps");
      g2.writePS( f );
    }
    break;
  }
}

int main( int argc, char **argv )
{
  // initialize screen output for both scenes
  g.initGL( &argc, argv, "Test", 640, 640 );
  g2.initGL( &argc, argv, "Test 2", 480, 480 );
  
  // register keyboard callbacks for both scenes
  g.setKeyboardHandler( gkeyboard );
  g2.setKeyboardHandler( g2keyboard );

  // create a drawable point object at (1,1.5) with a
  //  diameter of 8
  drawablePoint p(2.0,1.5,8.0);

  // create a drawable line segment from (1,1) to (9,1)
  //  with a custom dash pattern
  drawableSegment s( Point(1.0,1.0),Point(9.0,1.0),
		     drawableSegment::getDefaultLineWidth(), 0xfcfc );
  // make the segment red
  s.setColor( Color(1.0, 0, 0) );
  
  // create a line strip (a series of vertices)
  drawableLineStrip ls;
  ls.setColor( Color(0, 1.0, 0) ); // green
  ls.push_back( Point(5.0,3.0) );  // first point
  ls.push_back( Point(6.0,3.5) );  // second point
  ls.push_back( Point(5.0,4.0) );  // third point
  
  // create a polygon (solid border, filled, border width of 3)
  drawablePolygon o(true,3.0f,SOLID);
  o.setColor( Color(0, 0, 1.0) ); // blue filling
  o.push_back( Point(7.0,7.0) );  // first point
  o.push_back( Point(9.0,9.0) );  // second point
  o.push_back( Point(9.5,7.0) );  // third point
  o.push_back( Point(8.0,6.5) );  // fourth point

  // create a non-convex polygon with holes
  drawablePolygon nc;
  nc.push_back( Point(1.0,2.0) );
  nc.push_back( Point(2.5,3.0) );
  nc.push_back( Point(4.0,2.0) );
  nc.push_back( Point(3.5,5.0) );
  nc.push_back( Point(4.0,6.0) );
  nc.push_back( Point(1.0,5.0) );
  nc.setColor( "blue" );
  nc.setLineWidth( 3.0 );
  nc.setBorderColor( "green" );

  // create an ellipse centered at (2,8), with an orientation
  //  of -1.2 radians, a major radius of 2.0 and a minor radius
  //  of 1.0
  drawableEllipse e( Point(2.0,8.0),-1.2,2.0,1.0 );
  e.setColor( Color(1.0, 1.0, 0) ); // yellow

  // create a circle centered at (8,2) with a radius of 1.5^2, a
  //  dashed border, filled, with a border width of 3
  drawableCircle c( Point(8.0,2.0),1.5,true,3.0f,DASHED );
  // set the circle's color to cyan, 50% transparent
  c.setColor( Color(0, 1.0, 1.0, 0.5) );

  // create a rectangle with a lower left corner at (5,5) and an
  //  upper right corner at (7,7), with a solid border, filled,
  //  and a border width of 5
  drawableRectangle r(Point(5.0,5.0),Point(7.0,7.0),true,5.0f,SOLID);
  // set the filling color to 50% gray
  r.setColor( Color("gray50") );
  // set the border color to red
  r.setBorderColor( Color("red") );

  // create an arc centered at (5,8.5), with a squared radius of
  //  1.0, arcing (counterclockwise) from PI/2 to PI
  drawableArc a(Point(5.0,8.5),1.0,M_PI / 2.0,M_PI,3.0f,SOLID);
  a.setColor( "green" );

  // create a wedge centered at (5,8.5) with a squared radius of
  //  1.0, arcing (counterclockwise) from PI to 2PI
  drawableWedge w(Point(5.0,8.5),1.0,M_PI,M_PI * 2.0,true,2.0f,SOLID);
  w.setColor( "magenta" );

  // create a new objectList so that we can add all of the
  //  objects to the first scene
  objectList olist;
  olist.push_back( &p );  // add point
  olist.push_back( &s );  // add line segment
  olist.push_back( &ls ); // add line strip
  olist.push_back( &o );  // add polygon
  olist.push_back( &nc ); // add non-convex polygon with hole
  olist.push_back( &e );  // add ellipse
  olist.push_back( &c );  // add circle
  olist.push_back( &r );  // add rectangle
  olist.push_back( &a );  // add arc
  olist.push_back( &w );  // add wedge
  g.addObjectList( olist );  // add the objectList to the scene

  // create a new objectList to add objects to the second scene
  objectList olist2;
  olist2.push_back( &r ); // add rectangle
  g2.addObjectList( olist2 ); // add the objectList to the scene

  // run the main event loop, drawing everything on the screen
  g.runEventLoop();
  return 0;
};



