/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: drawable.h,v 1.13 2003/05/27 19:14:17 beevek Exp $
*/

#ifndef _DRAWABLE_H
#define _DRAWABLE_H

////////////////////////////////////////////////////////////////
//
//  Base classes for various kinds of drawableObjects
//  
////////////////////////////////////////////////////////////////

#include <dolt-lite/geometry.h>
#include <dolt-lite/color.h>
#include <stdio.h>

#ifdef _WIN32
  // get rid of annoying warnings in MSVC++ about uncalled
  // constructors
  #pragma warning( disable : 4930 )
#endif // _WIN32

namespace dolt {

class Graphics;
class psFile;

// base class for all drawable objects
class drawableObject
{
protected:
  Color col;
  static Color defaultColor;

  // set openGL's current color to the color of this object
  void setGLColor() const;
  void setPSColor( psFile &out ) const;

public:
  virtual ~drawableObject() {}
  // set default color for all objects
  static inline void setDefaultObjectColor( const Color &c )
  {
    defaultColor = c ;
  }
  static inline Color getDefaultObjectColor()
  {
    return defaultColor;
  }

  // all objects have a 'color'
  inline void setColor( const Color &c ) { col = c; }
  inline void setAlpha( float a ) { col.setAlpha( a ); }

  inline const Color & getColor() const { return col; }

  // all derived classes must define these (though of course
  //  they can be left empty).  drawGL implements a method
  //  for drawing the drawableObject in OpenGL to the output
  //  device described by <out>; drawPS implements a method
  //  for drawing the drawableObject to the postscript file
  //  described by <ps> (and using the information and
  //  methods provided by <out>)
  virtual void drawGL( const Graphics &out ) const = 0;
  virtual void drawPS( const Graphics &out, psFile &ps,
		       const Bbox &bounds, const double &scale ) const = 0;
};


////////////////
//
//  Line graphics
//
//  (segments, segment chains, polygons, ellipses, circles, etc)
//  

typedef unsigned short lineStyle;
const lineStyle SOLID = 0xffff;
const lineStyle DOTTED = 0xcccc;
const lineStyle DASHED = 0xff00;

// base class for line-based objects (e.g. line segment,
//  line strip, polygons, ellipses, circles)
class drawableLineObject : public drawableObject
{
protected:
  static float defaultLineWidth;
  static lineStyle defaultLineStyle;

  float lwidth;
  lineStyle lstyle;

  void setupGLLineDrawing() const;
  void setupPSLineDrawing( psFile &out ) const;
  void endGLLineDrawing() const;
  void endPSLineDrawing( psFile &out ) const;

  // put constructor here so that this class can't be instantiated
  drawableLineObject( float lineWidth = defaultLineWidth,
		      lineStyle style = defaultLineStyle ) 
    : lwidth(lineWidth), lstyle(style) {}
public:
  // set default drawing width of all lined objects
  static inline void setDefaultLineWidth( float width )
  {
    defaultLineWidth = width;
  }
  static inline float getDefaultLineWidth()
  {
    return defaultLineWidth;
  }

  static inline void setDefaultLineStyle( lineStyle style )
  {
    defaultLineStyle = style;
  }
  static inline lineStyle getDefaultLineStyle()
  {
    return defaultLineStyle;
  }

  inline void setLineWidth( float width ) { lwidth = width; }
  inline void setLineStyle( lineStyle style ) { lstyle = style; }

  inline float getLineWidth() const { return lwidth; }
  inline lineStyle getLineStyle() const { return lstyle; }
};

////////////////
//
//  Enclosed objects (e.g. figures that enclose area that may be "filled")
//
//  since enclosed objects can have borders, this is derived from
//  drawableLineObject
//  
class drawableEnclosedObject : public drawableLineObject
{
protected:
  static Color defaultBorderColor;
  // defaultLineWidth here, like that in drawableLineObject,
  //  refers to the line width for the object (in this
  //  case, specifically for the border); because the
  //  default line width for enclosed objects should be
  //  able to be different from that for line objects,
  //  we redefine this here
  static float defaultLineWidth;
  static bool defaultFilled;

  bool objectFilled;
  Color borderColor;
  // set the opengl current color to the color of the
  //  object's border
  void setGLBorderColor() const;
  void setPSBorderColor( psFile &out ) const;
  
  // put constructor here so that this class can't be instantiated
  drawableEnclosedObject( bool filled = defaultFilled,
			  float lineWidth = defaultLineWidth,
			  lineStyle style = defaultLineStyle )
  : drawableLineObject(lineWidth, style), objectFilled(filled),
    borderColor(defaultBorderColor) {}
public:
  static inline void setDefaultBorderColor( const Color &c )
  {
    defaultBorderColor = c;
  }
  static inline void setDefaultLineWidth( float width )
  {
    defaultLineWidth = width;
  }
  static inline Color getDefaultBorderColor()
  {
    return defaultBorderColor;
  }
  static inline float getDefaultLineWidth()
  {
    return defaultLineWidth;
  }

  inline void setFilled( bool filled ) { objectFilled = filled; }
  inline void setBorderColor( const Color &c ) { borderColor = c; }

  inline bool getFilled() const { return objectFilled; }
  inline const Color & getBorderColor() const { return borderColor; }
};

////////////////////////////////////////////////////////////////
//
//  A number of simple drawable objects of different types
//    
//  The advantage in implementing drawing of objects in this manner is
//  that it makes the graphics largely independent of the geometry
//  library being used, as long as the geometry library provides an
//  interface similar to that in geometry.h.
//  
////////////////////////////////////////////////////////////////

////////////////
// Points
// 
// FIXME: how to do symbols for point art?
class drawablePoint : public drawableObject, public Point
{
protected:
  float size;
  static float defaultSize;
public:
  drawablePoint() { drawablePoint(Point()); }
  drawablePoint( const double &px, const double &py, 
		 float pointSize = defaultSize,
		 const Color &color = defaultColor ) 
    : Point(px,py), size(pointSize)
  {
    setColor(color);
  }
  drawablePoint( const Point &p, float pointSize = defaultSize,
		 const Color &color = defaultColor )
    : Point(p), size(pointSize)
  {
    setColor( color );
  }

  // set default point drawing size for all points
  static inline void setDefaultSize( float s )
  {
    defaultSize = s;
  }
  static inline float getDefaultSize()
  {
    return defaultSize;
  }

  inline void setSize( float pointSize ) { size = pointSize; }

  void operator=( const Point &p ) { Point::operator=(p); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};


////////////////
// Line segments
// 
class drawableSegment : public drawableLineObject, public Segment
{
public:
  drawableSegment() { drawableSegment(Segment()); }
  drawableSegment( const Point &start, const Point &end,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style), Segment(start,end)
  {
    setColor(color);
  }
  drawableSegment( const Segment &s,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style), Segment(s)
  {
    setColor(color);
  }

  void operator=( const Segment &s ) { Segment::operator=(s); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
    const Bbox &bounds, const double &scale ) const;
};

////////////////
// chain of line segments
// 
class drawableLineStrip : public drawableLineObject, public lineStrip
{
public:
  drawableLineStrip( float lineWidth = defaultLineWidth,
		     lineStyle style = defaultLineStyle,
		     const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style)
  {
    setColor(color);
  }
  drawableLineStrip( const lineStrip &l,
		     float lineWidth = defaultLineWidth,
		     lineStyle style = defaultLineStyle,
		     const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style), lineStrip(l)
  {
    setColor(color);
  }

  void operator=( const lineStrip &l ) { lineStrip::operator=(l); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// polygons
// 
// NOTE: polygons in dolt-lite must be convex.  nonconvex polygons
// will look funny!  if you want nonconvex polygons or polygons with
// holes, install the full dolt library.
class drawablePolygon : public drawableEnclosedObject, public Polygon
{
public:
  drawablePolygon( bool filled = defaultFilled,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style)
  {
    setColor(color);
  }
  drawablePolygon( const Polygon &p,
		   bool filled = defaultFilled,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style), Polygon(p)
  {
    setColor(color);
  }

  void operator=( const Polygon &p ) { Polygon::operator=(p); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// ellipses
// 
class drawableEllipse : public drawableEnclosedObject, public Ellipse
{
public:
  drawableEllipse() { drawableEllipse(Ellipse()); }
  drawableEllipse( const Point &p, const double &orient,
		   const double &majorRadius, const double &minorRadius,
		   bool filled = defaultFilled,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style),
      Ellipse(p,orient,majorRadius,minorRadius)
  {
    setColor(color);
  }
  drawableEllipse( const Ellipse &e,
		   bool filled = defaultFilled,
		   float lineWidth = defaultLineWidth,
		   lineStyle style = defaultLineStyle,
		   const Color &color = defaultColor)
    : drawableEnclosedObject(filled,lineWidth,style), Ellipse(e)
  {
    setColor(color);
  }

  void operator=( const Ellipse &e ) { Ellipse::operator=(e); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// circles
// 
class drawableCircle : public drawableEnclosedObject, public Circle
{
public:
  drawableCircle() { drawableCircle(Circle()); }
  drawableCircle( const Point &center, const double &squaredRadius,
		  bool filled = defaultFilled,
		  float lineWidth = defaultLineWidth,
		  lineStyle style = defaultLineStyle,
		  const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style),
      Circle(center,squaredRadius)
  {
    setColor(color);
  }
  drawableCircle( const Circle &c,
		  bool filled = defaultFilled,
		  float lineWidth = defaultLineWidth,
		  lineStyle style = defaultLineStyle,
		  const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style), Circle(c)
  {
    setColor(color);
  }

  void operator=( const Circle &c ) { Circle::operator=(c); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// rectangle (axis aligned)
// 
class drawableRectangle : public drawableEnclosedObject, public Rectangle
{
public:
  drawableRectangle() { drawableRectangle(Rectangle()); }
  drawableRectangle( const Point &center,
		     const double &width, const double &height,
		     bool filled = defaultFilled,
		     float lineWidth = defaultLineWidth,
		     lineStyle style = defaultLineStyle,
		     const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style),
      Rectangle(center,width,height)
  {
    setColor(color);
  }
  drawableRectangle( const Point &lowerLeft, const Point &upperRight,
		     bool filled = defaultFilled,
		     float lineWidth = defaultLineWidth,
		     lineStyle style = defaultLineStyle,
		     const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style),
      Rectangle(lowerLeft,upperRight)
  {
    setColor(color);
  }
  drawableRectangle( const Rectangle &b,
		     bool filled = defaultFilled,
		     float lineWidth = defaultLineWidth,
		     lineStyle style = defaultLineStyle,
		     const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style), Rectangle(b)
  {
    setColor(color);
  }

  void operator=( const Rectangle &r ) { Rectangle::operator=(r); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// arcs (note that these are just arc lines, not wedges)
//
class drawableArc : public drawableLineObject, public Arc
{
public:
  drawableArc() { drawableArc(Arc()); }
  drawableArc( const Point &center, const double &squaredRad,
	       const double &minAngle, const double &maxAngle,
	       float lineWidth = defaultLineWidth,
	       lineStyle style = defaultLineStyle,
	       const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style),
      Arc(center,squaredRad,minAngle,maxAngle)
  {
    setColor(color);
  }
  drawableArc( const Arc &a,
	       float lineWidth = defaultLineWidth,
	       lineStyle style = defaultLineStyle,
	       const Color &color = defaultColor )
    : drawableLineObject(lineWidth,style), Arc(a)
  {
    setColor(color);
  }

  void operator=( const Arc &a ) { Arc::operator=(a); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// wedges (filled arcs)
//
class drawableWedge : public drawableEnclosedObject, public Arc
{
public:
  drawableWedge() { drawableWedge(Arc()); }
  drawableWedge( const Arc &a,
		 bool filled = defaultFilled,
		 float lineWidth = defaultLineWidth,
		 lineStyle style = defaultLineStyle,
		 const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style), Arc(a)
  {
    setColor(color);
  }
  drawableWedge( const Point &center, const double &squaredRad,
		 const double &minAngle, const double &maxAngle,
		 bool filled = defaultFilled,
		 float lineWidth = defaultLineWidth,
		 lineStyle style = defaultLineStyle,
		 const Color &color = defaultColor )
    : drawableEnclosedObject(filled,lineWidth,style),
      Arc(center,squaredRad,minAngle,maxAngle)
  {
    setColor(color);
  }

  void operator=( const Arc &a ) { Arc::operator=(a); }

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

////////////////
// images (raw pixel data)
//
class drawableImage : public drawableObject
{
public:
  // BPP1 can be used for simple grayscale bitmaps; note that
  //  3 bytes per pixel are always read and/or written from/to
  //  PPM files
  typedef enum { BPP1 = 1, BPP3 = 3 } bytespp;
protected:
  // location of image in world coordinates
  Bbox worldLoc;
  unsigned int width, height; // width and height
  unsigned char *buffer;
  bytespp bpp;

  bool fail( FILE *in );
  void destroy();
  bool readAscii( FILE *in, unsigned int maxval );
  bool readBinary( FILE *in );
public:
  drawableImage()
    : worldLoc(0.0,0.0,0.0,0.0), width(0), height(0),
      buffer(0), bpp(BPP3) {}
  drawableImage( const Bbox &worldLocation, unsigned int w,
		 unsigned int h, bytespp bytes = BPP3 );
  drawableImage( const drawableImage &i );
  ~drawableImage();

  inline void setWorldLocation( const Bbox &worldLocation )
  {
    worldLoc = worldLocation;
  }
  inline unsigned int getWidth() const { return width; }
  inline unsigned int getHeight() const { return height; }
  inline unsigned char * getBuffer() { return buffer; }

  void setPixel( unsigned int x, unsigned int y, const unsigned char *col );
  const unsigned char * const getPixel( unsigned int x, unsigned int y ) const;

  // load/save a ppm file
  bool loadPPM( const char *filename );
  bool savePPM( const char *filename, bool binaryFile = true ) const;

  void drawGL( const Graphics &out ) const;
  void drawPS( const Graphics &out, psFile &ps,
	       const Bbox &bounds, const double &scale ) const;
};

} // end of namespace dolt

#endif // _DRAWABLE_H










