/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: graphics.h,v 1.9 2004/01/27 02:37:07 whuang Exp $
*/

#ifndef _GRAPHICS_H
#define _GRAPHICS_H

#ifdef _WIN32
  // get rid of annoying stl warnings in MSVC++ about
  // truncated debug symbols (because of long stl names)
  #pragma warning( disable : 4786 )
#endif // _WIN32

#include <GL/glut.h>
#include <list>
#include <dolt-lite/drawable.h>
#include <dolt-lite/geometry.h>
#include <dolt-lite/postscript.h>
#include <dolt-lite/color.h>

namespace dolt {

class objectList : public std::list<drawableObject *> 
{
private:
  bool _visible;
public:
  objectList() : _visible(true) {}

  bool visible() const { return _visible; }
  void show() { _visible = true; }
  void hide() { _visible = false; }
  void toggle_visibility() { _visible = !_visible; }
};

class Graphics
{
private:
  // the currently selected Graphics instance for drawing to the screen
  static Graphics *currentInstance;
  static bool glutInitialized;

  // tell glut to use the settings for this instance of Graphics (for
  // initialization and switching between windows)
  void setupWindow();

  // list of object lists to draw
  std::list<const objectList *> objectLists;

  // bounding box for the world (in world coordinates)
  Bbox worldbox;

  // zoomPercent: percentage zooming of scene from original size
  //  i.e. if zoomPercent == 0.5 (50%), you're zoomed out so that
  //  the scene appears to be 50% of its full size, and if
  //  zoomPercent == 2.0 (200%), you're zoomed in so that the
  //  scene appears to be 200% of its full size; note that the
  //  point around which zooming takes place is the center of
  //  the scene
  float zoomPercent;
  // panPercent{X,Y}: percentage of x and y dimensions of
  //  original scene that the scene is panned away from center.
  //  i.e. if panPercent{X,Y} == (0.0,-1.0) the scene will be
  //  centered in the x direction and scene.ymax - scene.ymin
  //  to the left in the y direction
  float panPercentX, panPercentY;
  
  // width and height (in pixels) of window for screen output
  int pixW, pixH;
  Color bgColor;

  // window title
  char *title;

  // pointers to user keyboard/mouse/idle handlers
  void (*userKeyboard)( unsigned char key );
  void (*userMouse)( int button, int state, int x, int y );
  void (*userMotion)( int x, int y );
  void (*userIdle)();
  void (*animateCallback)();
  // default handlers for keyboard events
  static void keyboard( unsigned char key, int x, int y );
  static void specialKeyboard( int key, int x, int y );

  // handle window reshaping (keep everything centered and
  //  keep the proper aspect ratios)
  static void reshape( int w, int h );

  // draw everything
  static void display();

public:
  Graphics();
  Graphics( const Bbox &world );
  ~Graphics();

  // initialize glut
  void initGL( int *argc, char **argv, const char *name = 0,
	       int winWidth = 750, int winHeight = 0,
	       int winPosX = 100, int winPosY = 100 );
  
  // add/remove lists of objects to be drawn
  void addObjectList( const objectList &olist );
  void delObjectList( const objectList &olist );

  // draw objects in the given Graphics instance to the screen
  static void setCurrentScene( Graphics &g );

  // run main drawing loop
  static void runEventLoop();

  // request that the current screen be rendered in postscript to
  // <out>, with bounding box in the document described by bounds
  void writePS( psFile &out, const Bbox &bounds = Bbox(0.5,0.5,8.0,10.5) ) const;
  
  // request that the current screen be dumped (in binary ppm
  //  format) to <filename>; false on failure, true on success
  bool writePPM( const char *filename ) const;

  inline void setWorldBounds( const Bbox &world ) { worldbox = world; }
  void setBackgroundColor( const Color &c );

  // install user-defined handlers
  void setKeyboardHandler( void (*kb)( unsigned char key ) );
  void setMouseHandler( void (*mouse)( int button, int state, int x, int y ) );
  void setMotionHandler( void (*motion)( int x, int y ) );
  void setIdleFunction( void (*idle)( void ) );
  void setAnimateCallback( void (*processing)( void ) );
  
  // print default control (pan/zoom) key bindings to stdout
  void printControlKeyBindings() const;

  // zoom: percentage zoomed, 100% (1.0) is normal
  //  zoom, < 100% zooms out, > 100% zooms in
  void zoom( float newZoom = 1.0 );
  // pan: < 0%x moves left, > 0%x moves right,
  //  < 0%y moves down, > 0%y moves up
  void pan( float panX, float panY );

  void pixelsPerUnit( float &ppu ) const;
  inline double getWidth() const { return worldbox.xmax() - worldbox.xmin(); }
  inline double getHeight() const { return worldbox.ymax() - worldbox.ymin(); }
  inline int getPixWidth() const { return pixW; }
  inline int getPixHeight() const { return pixH; }

  // map <p> from world coordinates (taking into account
  //  zooming and panning) to the world boundary
  //  described by <bounds>
  Point mapPoint( const Bbox &bounds, const double &scale,
		  const Point &p ) const;
};

} // end namespace dolt

#endif // _GRAPHICS_H




