/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: color.cpp,v 1.4 2003/05/27 18:20:46 beevek Exp $
*/

#include <dolt-lite.h>
#include <string>

using namespace dolt;

// mapping names (strings) to colors; note that
//  since the constructor is private, only Color
//  (a friend) can instantiate this class
class dolt::colorMap
{
  friend class Color;
private:
  // for map lookup
  struct ltstring
  {
    bool operator()( const char *s0, const char *s1 ) const
    {
      return strcmp( s0, s1 ) < 0;
    }
  };
  std::map<const char *,Color,ltstring> colors;

  // initialize default colors in the constructor
  colorMap();
  // lookup of colors
  const Color & operator[]( const char *name );
  void add( const char *name, const Color &c );
};

colorMap Color::colors;

Color::Color( const char *name )
{
  *this = colors[name];
}

Color::Color( const char *name, float a )
{
  *this = colors[name];
  alpha = a;
}

inline void Color::setRGB( float r, float g, float b )
{
  red = r;
  green = g;
  blue = b;
}

void Color::nameColor( const char *name )
{
  colors.add( name, *this );
}

inline void eatwhite( std::istream &in )
{
  while( strchr( " \t", in.peek() ) )
    in.get();
}

std::istream & dolt::operator>>( std::istream &in, Color &c )
{
  // we can either read a string (a color name) or a
  //  series of 3 whitespace-separated floats between
  //  0.0 and 1.0
  // assume there is no whitespace at the front of the
  //  stream

  if( strchr( "1234567890.", in.peek() ) ) {
    // starts with a digit or a decimal; we require
    //  that this be a color defined by 3 floats since
    //  names can't start with digits or decimals
    float r, g, b;
    in >> r;
    eatwhite( in );
    in >> g;
    eatwhite( in );
    in >> b;
    c = Color(r,g,b);
  } else {
    // otherwise we assume it is a name
    std::string name;
    in >> name;
    c = Color(name.c_str());
  }
  return in;
}

const Color & colorMap::operator[]( const char *name )
{
  return colors[name];
}

void colorMap::add( const char *name, const Color &c )
{
  colors[name] = c;
}

colorMap::colorMap()
{
  // generate default colors
  colors["black"] = Color( 0.00f, 0.00f, 0.00f );
  colors["navy"] = Color( 0.00f, 0.00f, 0.50f );
  colors["darkblue"] = Color( 0.00f, 0.00f, 0.55f );
  colors["mediumblue"] = Color( 0.00f, 0.00f, 0.80f );
  colors["blue"] = Color( 0.00f, 0.00f, 1.00f );
  colors["darkgreen"] = Color( 0.00f, 0.39f, 0.00f );
  colors["green"] = Color( 0.00f, 0.50f, 0.00f );
  colors["teal"] = Color( 0.00f, 0.50f, 0.50f );
  colors["darkcyan"] = Color( 0.00f, 0.55f, 0.55f );
  colors["deepskyblue"] = Color( 0.00f, 0.75f, 1.00f );
  colors["darkturquoise"] = Color( 0.00f, 0.81f, 0.82f );
  colors["mediumspringgreen"] = Color( 0.00f, 0.98f, 0.60f );
  colors["lime"] = Color( 0.00f, 1.00f, 0.00f );
  colors["springgreen"] = Color( 0.00f, 1.00f, 0.50f );
  colors["aqua"] = Color( 0.00f, 1.00f, 1.00f );
  colors["cyan"] = Color( 0.00f, 1.00f, 1.00f );
  colors["royalblue"] = Color( 0.02f, 0.09f, 0.56f );
  colors["midnightblue"] = Color( 0.10f, 0.10f, 0.44f );
  colors["dodgerblue"] = Color( 0.12f, 0.56f, 1.00f );
  colors["lightseagreen"] = Color( 0.13f, 0.70f, 0.67f );
  colors["forestgreen"] = Color( 0.13f, 0.55f, 0.13f );
  colors["seagreen"] = Color( 0.18f, 0.55f, 0.34f );
  colors["darkslategray"] = Color( 0.18f, 0.31f, 0.31f );
  colors["limegreen"] = Color( 0.20f, 0.80f, 0.20f );
  colors["mediumseagreen"] = Color( 0.24f, 0.70f, 0.44f );
  colors["turquoise"] = Color( 0.25f, 0.88f, 0.82f );
  colors["royalblue"] = Color( 0.25f, 0.41f, 0.88f );
  colors["steelblue"] = Color( 0.27f, 0.51f, 0.71f );
  colors["darkslateblue"] = Color( 0.28f, 0.24f, 0.55f );
  colors["mediumturquoise"] = Color( 0.28f, 0.82f, 0.80f );
  colors["indigo"] = Color( 0.29f, 0.00f, 0.51f );
  colors["darkolivegreen"] = Color( 0.33f, 0.42f, 0.18f );
  colors["cadetblue"] = Color( 0.37f, 0.62f, 0.63f );
  colors["cornflowerblue"] = Color( 0.39f, 0.58f, 0.93f );
  colors["mediumaquamarine"] = Color( 0.40f, 0.80f, 0.67f );
  colors["dimgray"] = Color( 0.41f, 0.41f, 0.41f );
  colors["slateblue"] = Color( 0.42f, 0.35f, 0.80f );
  colors["olivedrab"] = Color( 0.42f, 0.56f, 0.14f );
  colors["slategray"] = Color( 0.44f, 0.50f, 0.56f );
  colors["lightslategray"] = Color( 0.47f, 0.53f, 0.60f );
  colors["mediumslateblue"] = Color( 0.48f, 0.41f, 0.93f );
  colors["lawngreen"] = Color( 0.49f, 0.99f, 0.00f );
  colors["chartreuse"] = Color( 0.50f, 1.00f, 0.00f );
  colors["aquamarine"] = Color( 0.50f, 1.00f, 0.83f );
  colors["maroon"] = Color( 0.50f, 0.00f, 0.00f );
  colors["purple"] = Color( 0.50f, 0.00f, 0.50f );
  colors["olive"] = Color( 0.50f, 0.50f, 0.00f );
  colors["gray"] = Color( 0.50f, 0.50f, 0.50f );
  colors["skyblue"] = Color( 0.53f, 0.81f, 0.92f );
  colors["lightskyblue"] = Color( 0.53f, 0.81f, 0.98f );
  colors["blueviolet"] = Color( 0.54f, 0.17f, 0.89f );
  colors["darkred"] = Color( 0.55f, 0.00f, 0.00f );
  colors["darkmagenta"] = Color( 0.55f, 0.00f, 0.55f );
  colors["saddlebrown"] = Color( 0.55f, 0.27f, 0.07f );
  colors["darkseagreen"] = Color( 0.56f, 0.74f, 0.56f );
  colors["lightgreen"] = Color( 0.56f, 0.93f, 0.56f );
  colors["mediumpurple"] = Color( 0.58f, 0.44f, 0.86f );
  colors["darkviolet"] = Color( 0.58f, 0.00f, 0.83f );
  colors["palegreen"] = Color( 0.60f, 0.98f, 0.60f );
  colors["darkorchid"] = Color( 0.60f, 0.20f, 0.80f );
  colors["yellowgreen"] = Color( 0.60f, 0.80f, 0.20f );
  colors["sienna"] = Color( 0.63f, 0.32f, 0.18f );
  colors["brown"] = Color( 0.65f, 0.16f, 0.16f );
  colors["darkgray"] = Color( 0.66f, 0.66f, 0.66f );
  colors["lightblue"] = Color( 0.68f, 0.85f, 0.90f );
  colors["greenyellow"] = Color( 0.68f, 1.00f, 0.18f );
  colors["paleturquoise"] = Color( 0.69f, 0.93f, 0.93f );
  colors["lightsteelblue"] = Color( 0.69f, 0.77f, 0.87f );
  colors["powderblue"] = Color( 0.69f, 0.88f, 0.90f );
  colors["firebrick"] = Color( 0.70f, 0.13f, 0.13f );
  colors["darkgoldenrod"] = Color( 0.72f, 0.53f, 0.04f );
  colors["mediumorchid"] = Color( 0.73f, 0.33f, 0.83f );
  colors["rosybrown"] = Color( 0.74f, 0.56f, 0.56f );
  colors["darkkhaki"] = Color( 0.74f, 0.72f, 0.42f );
  colors["silver"] = Color( 0.75f, 0.75f, 0.75f );
  colors["mediumvioletred"] = Color( 0.78f, 0.08f, 0.52f );
  colors["indianred"] = Color( 0.80f, 0.36f, 0.36f );
  colors["peru"] = Color( 0.80f, 0.52f, 0.25f );
  colors["chocolate"] = Color( 0.82f, 0.41f, 0.12f );
  colors["tan"] = Color( 0.82f, 0.71f, 0.55f );
  colors["lightgrey"] = Color( 0.83f, 0.83f, 0.83f );
  colors["thistle"] = Color( 0.85f, 0.75f, 0.85f );
  colors["orchid"] = Color( 0.85f, 0.44f, 0.84f );
  colors["goldenrod"] = Color( 0.85f, 0.65f, 0.13f );
  colors["palevioletred"] = Color( 0.86f, 0.44f, 0.58f );
  colors["crimson"] = Color( 0.86f, 0.08f, 0.24f );
  colors["gainsboro"] = Color( 0.86f, 0.86f, 0.86f );
  colors["plum"] = Color( 0.87f, 0.63f, 0.87f );
  colors["burlywood"] = Color( 0.87f, 0.72f, 0.53f );
  colors["lightcyan"] = Color( 0.88f, 1.00f, 1.00f );
  colors["lavender"] = Color( 0.90f, 0.90f, 0.98f );
  colors["darksalmon"] = Color( 0.91f, 0.59f, 0.48f );
  colors["violet"] = Color( 0.93f, 0.51f, 0.93f );
  colors["palegoldenrod"] = Color( 0.93f, 0.91f, 0.67f );
  colors["lightcoral"] = Color( 0.94f, 0.50f, 0.50f );
  colors["khaki"] = Color( 0.94f, 0.90f, 0.55f );
  colors["aliceblue"] = Color( 0.94f, 0.97f, 1.00f );
  colors["honeydew"] = Color( 0.94f, 1.00f, 0.94f );
  colors["azure"] = Color( 0.94f, 1.00f, 1.00f );
  colors["sandybrown"] = Color( 0.96f, 0.64f, 0.38f );
  colors["wheat"] = Color( 0.96f, 0.87f, 0.70f );
  colors["beige"] = Color( 0.96f, 0.96f, 0.86f );
  colors["whitesmoke"] = Color( 0.96f, 0.96f, 0.96f );
  colors["mintcream"] = Color( 0.96f, 1.00f, 0.98f );
  colors["ghostwhite"] = Color( 0.97f, 0.97f, 1.00f );
  colors["salmon"] = Color( 0.98f, 0.50f, 0.45f );
  colors["antiquewhite"] = Color( 0.98f, 0.92f, 0.84f );
  colors["linen"] = Color( 0.98f, 0.94f, 0.90f );
  colors["lightgoldenrodyellow"] = Color( 0.98f, 0.98f, 0.82f );
  colors["oldlace"] = Color( 0.99f, 0.96f, 0.90f );
  colors["red"] = Color( 1.00f, 0.00f, 0.00f );
  colors["fuchsia"] = Color( 1.00f, 0.00f, 1.00f );
  colors["magenta"] = Color( 1.00f, 0.00f, 1.00f );
  colors["deeppink"] = Color( 1.00f, 0.08f, 0.58f );
  colors["orangered"] = Color( 1.00f, 0.27f, 0.00f );
  colors["tomato"] = Color( 1.00f, 0.39f, 0.28f );
  colors["hotpink"] = Color( 1.00f, 0.41f, 0.71f );
  colors["coral"] = Color( 1.00f, 0.50f, 0.31f );
  colors["darkorange"] = Color( 1.00f, 0.55f, 0.00f );
  colors["lightsalmon"] = Color( 1.00f, 0.63f, 0.48f );
  colors["orange"] = Color( 1.00f, 0.65f, 0.00f );
  colors["lightpink"] = Color( 1.00f, 0.71f, 0.76f );
  colors["pink"] = Color( 1.00f, 0.75f, 0.80f );
  colors["gold"] = Color( 1.00f, 0.84f, 0.00f );
  colors["peachpuff"] = Color( 1.00f, 0.85f, 0.73f );
  colors["navajowhite"] = Color( 1.00f, 0.87f, 0.68f );
  colors["moccasin"] = Color( 1.00f, 0.89f, 0.71f );
  colors["bisque"] = Color( 1.00f, 0.89f, 0.77f );
  colors["mistyrose"] = Color( 1.00f, 0.89f, 0.88f );
  colors["blanchedalmond"] = Color( 1.00f, 0.92f, 0.80f );
  colors["papayawhip"] = Color( 1.00f, 0.94f, 0.84f );
  colors["lavenderblush"] = Color( 1.00f, 0.94f, 0.96f );
  colors["seashell"] = Color( 1.00f, 0.96f, 0.93f );
  colors["cornsilk"] = Color( 1.00f, 0.97f, 0.86f );
  colors["lemonchiffon"] = Color( 1.00f, 0.98f, 0.80f );
  colors["floralwhite"] = Color( 1.00f, 0.98f, 0.94f );
  colors["snow"] = Color( 1.00f, 0.98f, 0.98f );
  colors["yellow"] = Color( 1.00f, 1.00f, 0.00f );
  colors["lightyellow"] = Color( 1.00f, 1.00f, 0.88f );
  colors["ivory"] = Color( 1.00f, 1.00f, 0.94f );
  colors["white"] = Color( 1.00f, 1.00f, 1.00f );
  colors["gray5"] = Color( 0.05f, 0.05f, 0.05f );
  colors["gray10"] = Color( 0.1f, 0.1f, 0.1f );
  colors["gray15"] = Color( 0.15f, 0.15f, 0.15f );
  colors["gray20"] = Color( 0.2f, 0.2f, 0.2f );
  colors["gray25"] = Color( 0.25f, 0.25f, 0.25f );
  colors["gray30"] = Color( 0.3f, 0.3f, 0.3f );
  colors["gray35"] = Color( 0.35f, 0.35f, 0.35f );
  colors["gray40"] = Color( 0.4f, 0.4f, 0.4f );
  colors["gray45"] = Color( 0.45f, 0.45f, 0.45f );
  colors["gray50"] = Color( 0.5f, 0.5f, 0.5f );
  colors["gray55"] = Color( 0.55f, 0.55f, 0.55f );
  colors["gray60"] = Color( 0.6f, 0.6f, 0.6f );
  colors["gray65"] = Color( 0.65f, 0.65f, 0.65f );
  colors["gray70"] = Color( 0.7f, 0.7f, 0.7f );
  colors["gray75"] = Color( 0.75f, 0.75f, 0.75f );
  colors["gray80"] = Color( 0.8f, 0.8f, 0.8f );
  colors["gray85"] = Color( 0.85f, 0.85f, 0.85f );
  colors["gray90"] = Color( 0.9f, 0.9f, 0.9f );
  colors["gray95"] = Color( 0.95f, 0.95f, 0.95f );
}
