/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: drawable.cpp,v 1.14 2004/01/13 16:02:17 whuang Exp $
*/

#include <GL/glut.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <dolt-lite.h>

using namespace dolt;

#if defined(_WIN32) || defined(_MSC_VER)
  namespace _W { // prevent name conflicts
  #include <windows.h>
  #undef max // so we can use numeric_limits<int>::max()
  };
#endif

#ifndef M_PI
  #define M_PI 3.1415926535897932384626433833
#endif
const double degreesToRadians = M_PI / 180.0;
const double radiansToDegrees = 180.0 / M_PI;

// defaults for object colors, sizes, etc.
Color drawableObject::defaultColor = Color(0.0,0.0,0.0);

float drawablePoint::defaultSize = 1.0f;

float drawableLineObject::defaultLineWidth = 1.0f;
lineStyle drawableLineObject::defaultLineStyle = SOLID;

Color drawableEnclosedObject::defaultBorderColor = Color(0.0,0.0,0.0);
float drawableEnclosedObject::defaultLineWidth = 0.0f;
bool drawableEnclosedObject::defaultFilled = true;


// drawableObject (base class) methods

void drawableObject::setGLColor() const
{
  // do blending here? (e.g. glEnable(GL_BLEND))
  glColor4f( col.r(), col.g(), col.b(), col.a() );
}

void drawableObject::setPSColor( psFile &out ) const
{
  out << col.r() << ' ' << col.g() << ' ' << col.b()
    << " setrgbcolor" << std::endl;
}

// drawablePoint methods

void drawablePoint::drawGL( const Graphics &out ) const
{
  setGLColor();
  glPointSize( size );
  glBegin( GL_POINTS );
  glVertex2d( x(), y() );
  glEnd();
}

void drawablePoint::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  ps.newpath();
  setPSColor( ps );
  Point p = ps.mapPoint( out, bounds, scale, *this );
  ps << std::setprecision(4) << p.x() << ' ' << p.y() << ' '
    << size / 2.0 << " 0 360 arc" << std::endl;
  ps << "gsave fill grestore stroke" << std::endl;
}

// drawableLineObject (base for all line objects) methods

void drawableLineObject::setupGLLineDrawing() const
{
  glLineWidth( lwidth );
  if( lstyle != SOLID ) {
    glLineStipple( 1, lstyle );
    glEnable( GL_LINE_STIPPLE );
  }
}

void drawableLineObject::setupPSLineDrawing( psFile &out ) const
{
  out << lwidth << " setlinewidth" << std::endl;
  out << "0 setlinejoin" << std::endl << "2 setlinecap" << std::endl;
  
  // this complicated stuff is to get the dash pattern
  //  out of lineStyle
  out << "[";
  int curCnt = 1;
  unsigned short curVal = lstyle & 0x0001, tmpVal;
  for( int i = 1; i < 16; i++ ) {
    tmpVal = (lstyle & (0x0001 << i)) >> i;
    if( tmpVal == curVal ) {
      curCnt++;
      continue;
    }
    out << curCnt << ' ';
    curCnt = 1;
    curVal = tmpVal;
  }
  if( curCnt < 16 )
    out << curCnt;
  out << "] 0 setdash" << std::endl;
}

void drawableLineObject::endGLLineDrawing() const
{
  glDisable( GL_LINE_STIPPLE );
}

void drawableLineObject::endPSLineDrawing( psFile &out ) const
{
  // nothing
}

// drawableEnclosedObject (base for enclosed objects)
//  methods

void drawableEnclosedObject::setGLBorderColor() const
{
  glColor4f( borderColor.r(), borderColor.g(),
    borderColor.b(), borderColor.a() );
}

void drawableEnclosedObject::setPSBorderColor( psFile &out) const
{
  out << borderColor.r() << ' ' << borderColor.g() << ' '
    << borderColor.b() << " setrgbcolor" << std::endl;
}

// drawableSegment methods

void drawableSegment::drawGL( const Graphics &out ) const
{
  if( lwidth <= 0.0 )
    return;
  setGLColor();
  setupGLLineDrawing();
  glBegin( GL_LINES );
  glVertex2d( start().x(), start().y() );
  glVertex2d( end().x(), end().y() );
  glEnd();
  endGLLineDrawing();
}

void drawableSegment::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( lwidth <= 0.0 )
    return;
  ps.newpath();
  setPSColor( ps );
  setupPSLineDrawing( ps );
  Point p;
  p = ps.mapPoint( out, bounds, scale, start() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " moveto" << std::endl;
  p = ps.mapPoint( out, bounds, scale, end() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " lineto" << std::endl << "stroke" << std::endl;
  endPSLineDrawing( ps );
}

// drawableLineStrip methods

void drawableLineStrip::drawGL( const Graphics &out ) const
{
  if( !size() )
    return;
  setGLColor();
  setupGLLineDrawing();
  lineStrip::const_iterator i;
  glBegin( GL_LINE_STRIP );
  for( i = begin(); i != end(); i++ )
    glVertex2d( i->x(), i->y() );
  glEnd();
  endGLLineDrawing();
}

void drawableLineStrip::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( !size() )
    return;
  ps.newpath();
  Point p;
  lineStrip::const_iterator i = begin();
  p = ps.mapPoint( out, bounds, scale, *i );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " moveto" << std::endl;
  for( ++i; i != end(); i++ ) {
    p = ps.mapPoint( out, bounds, scale, *i );
    ps << std::setprecision(4) << p.x() << ' ' << p.y()
      << " lineto" << std::endl;
  }
  setPSColor( ps );
  setupPSLineDrawing( ps );
  ps << "stroke" << std::endl;
  endPSLineDrawing( ps );
}

// drawablePolygon methods

void drawablePolygon::drawGL( const Graphics &out ) const
{
  if( !size() )
    return;
  Edge_const_iterator i = edges_begin();
  if( objectFilled ) { // draw the poly itself
    setGLColor();
    glBegin( GL_TRIANGLE_FAN );
    for(i = edges_begin(); i != edges_end(); i++)
      glVertex2d((*i).start().x(), (*i).start().y());
    glEnd();
  }
  if( lwidth > 0.0 ) { // draw the border
    setGLBorderColor();
    setupGLLineDrawing();
    Edge_const_iterator i;
    glBegin( GL_LINE_LOOP );
    for( i = edges_begin(); i != edges_end(); i++ )
      glVertex2d( (*i).start().x(), (*i).start().y() );
    glEnd();
    endGLLineDrawing();
  }
}

void drawablePolygon::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( !size() )
    return;
  ps.newpath();
  Point p;
  Edge_const_iterator i = edges_begin();
  p = ps.mapPoint( out, bounds, scale, (*i).start() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " moveto" << std::endl;
  for( ; i != edges_end(); i++ ) {
    p = ps.mapPoint( out, bounds, scale, (*i).end() );
    ps << std::setprecision(4) << p.x() << ' ' << p.y()
      << " lineto" << std::endl;
  }
  ps << "closepath" << std::endl;
  if( objectFilled ) {
    ps << "gsave" << std::endl;
    setPSColor( ps );
    ps << "eofill grestore" << std::endl;
  }
  if( lwidth > 0.0 ) {
    setPSBorderColor( ps );
    setupPSLineDrawing( ps );
    ps << "stroke" << std::endl;
    endPSLineDrawing( ps );
  }
}

// drawableEllipse methods

void drawableEllipse::drawGL( const Graphics &out ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;

  double maj_rad = getMajorRadius();
  double aspect_ratio = getMinorRadius() / maj_rad;
  Point loc = getPos();
  double orient = getOrientation();
  double cur_rot;

  // see drawableCircle::drawGL for info; we do this here with
  //  respect to major radius of the ellipse
  double inc = 360 /
    (4 + 356 * ((out.getWidth() > out.getHeight()) ?
		maj_rad / out.getWidth() : maj_rad / out.getHeight()));
  inc *= degreesToRadians;
  
  glPushMatrix();
  glTranslated( loc.x(), loc.y(), 0.0 );
  glRotated( orient * radiansToDegrees, 0, 0, 1 );
  glScaled( 1.0, aspect_ratio, 1.0 );

  if( objectFilled ) { // draw ellipse itself
    setGLColor();
    glBegin( GL_TRIANGLE_FAN );
    for( cur_rot = 0.0; cur_rot < 2.0 * M_PI; cur_rot += inc )
      glVertex2d( cos( cur_rot ) * maj_rad, sin( cur_rot ) * maj_rad );
    glVertex2d( cos( 0.0 ) * maj_rad, sin( 0.0 ) * maj_rad ); // close ellipse
    glEnd();
  }
  if( lwidth > 0.0 ) { // draw border
    setGLBorderColor();
    setupGLLineDrawing();
    glBegin( GL_LINE_LOOP );
    for( cur_rot = 0.0; cur_rot < 2.0 * M_PI; cur_rot += inc )
      glVertex2d( cos( cur_rot ) * maj_rad, sin( cur_rot ) * maj_rad );
    glEnd();
    endGLLineDrawing();
  }
  glPopMatrix();
}

void drawableEllipse::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;

  ps.newpath();
  Point p = ps.mapPoint( out, bounds, scale, getPos() );
  ps << std::setprecision(4) << "/mat matrix def" << std::endl
    << "/savemat mat currentmatrix def" << std::endl
    << p.x() << ' ' << p.y() << " translate" << std::endl
    << getOrientation() * radiansToDegrees << " rotate" << std::endl
    << ps.mapRange( scale, getMajorRadius() ) << ' '
    << ps.mapRange( scale, getMinorRadius() )
    << " scale" << std::endl << "0 0 1 0 360 arc" << std::endl
    << "savemat setmatrix" << std::endl;
  if( objectFilled ) {
    ps << "gsave" << std::endl;
    setPSColor( ps );
    ps << "fill grestore" << std::endl;
  }
  if( lwidth > 0.0 ) {
    setPSBorderColor( ps );
    setupPSLineDrawing( ps );
    ps << "stroke" << std::endl;
    endPSLineDrawing( ps );
  }
}

// drawableCircle methods

void drawableCircle::drawGL( const Graphics &out ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;

  double rad = (double)::sqrt( squared_radius() );
  Point loc = center();
  double cur_rot;
  // find a "good" angular increment based on the percentage
  //  of the largest world dimension of the radius
  //  FIXME: what is a better way to do LOD?
  //  FIXME: this doesn't take into account zooming
  double inc = 360 /
    (4 + 356 * ((out.getWidth() > out.getHeight()) ?
		rad / out.getWidth() : rad / out.getHeight()));
  inc *= degreesToRadians;

  glPushMatrix();
  glTranslated( loc.x(), loc.y(), 0.0 );

  if( objectFilled ) { // draw circle itself
    setGLColor();
    glVertex2d( 0.0, 0.0 );
    glBegin( GL_TRIANGLE_FAN );
    for( cur_rot = 0.0; cur_rot < 2.0 * M_PI; cur_rot += inc )
      glVertex2d( cos( cur_rot ) * rad, sin( cur_rot ) * rad );
    glVertex2d( cos( 0.0 ) * rad, sin( 0.0 ) * rad ); // close circle
    glEnd();
  }
  if( lwidth > 0.0 ) { // draw border
    setGLBorderColor();
    setupGLLineDrawing();
    glBegin( GL_LINE_LOOP );
    for( cur_rot = 0.0; cur_rot < 2.0 * M_PI; cur_rot += inc )
      glVertex2d( cos( cur_rot ) * rad, sin( cur_rot ) * rad );
    glEnd();
    endGLLineDrawing();
  }
  glPopMatrix();
}

void drawableCircle::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;

  ps.newpath();
  Point p = ps.mapPoint( out, bounds, scale, center() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y() << ' '
    << ps.mapRange( scale, ::sqrt( squared_radius() ) )
    << " 0 360 arc" << std::endl;
  if( objectFilled ) {
    ps << "gsave" << std::endl;
    setPSColor( ps );
    ps << "fill grestore" << std::endl;
  }
  if( lwidth > 0.0 ) {
    setPSBorderColor( ps );
    setupPSLineDrawing( ps );
    ps << "stroke" << std::endl;
    endPSLineDrawing( ps );
  }
}

// drawableRectangle methods

void drawableRectangle::drawGL( const Graphics &out ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;
  if( objectFilled ) { // draw rectangle itself
    setGLColor();
    glBegin( GL_QUADS );
    glVertex2d( xmin(), ymin() );
    glVertex2d( xmax(), ymin() );
    glVertex2d( xmax(), ymax() );
    glVertex2d( xmin(), ymax() );
    glEnd();
  }
  if( lwidth > 0.0 ) { // draw border
    setGLBorderColor();
    setupGLLineDrawing();
    glBegin( GL_LINE_LOOP );
    glVertex2d( xmin(), ymin() );
    glVertex2d( xmax(), ymin() );
    glVertex2d( xmax(), ymax() );
    glVertex2d( xmin(), ymax() );
    glEnd();
    endGLLineDrawing();
  }
}

void drawableRectangle::drawPS( const Graphics &out,
  psFile &ps, const Bbox &bounds, const double &scale ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;
  ps.newpath();
  Point p = ps.mapPoint( out, bounds, scale, Point(xmin(),ymin()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " moveto" << std::endl;
  p = ps.mapPoint( out, bounds, scale, Point(xmax(),ymin()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " lineto" << std::endl;
  p = ps.mapPoint( out, bounds, scale, Point(xmax(),ymax()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " lineto" << std::endl;
  p = ps.mapPoint( out, bounds, scale, Point(xmin(),ymax()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " lineto" << std::endl;
  p = ps.mapPoint( out, bounds, scale, Point(xmin(),ymin()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y()
    << " lineto" << std::endl << "closepath" << std::endl;
  if( objectFilled ) {
    ps << "gsave" << std::endl;
    setPSColor( ps );
    ps << "fill grestore" << std::endl;
  }
  if( lwidth > 0.0 ) {
    setPSBorderColor( ps );
    setupPSLineDrawing( ps );
    ps << "stroke" << std::endl;
    endPSLineDrawing( ps );
  }
}

void drawableArc::drawGL( const Graphics &out ) const
{
  if( lwidth <= 0.0 )
    return;

  glPushMatrix();
  glTranslated( center().x(), center().y(), 0.0 );
  setGLColor();
  setupGLLineDrawing();

  double rad = (double)::sqrt( squared_radius() );
  double cur_rot;

  // see drawableCircle::drawGL for explanation
  double inc = 360 /
    (4 + 356 * ((out.getWidth() > out.getHeight()) ?
		rad / out.getWidth() : rad / out.getHeight()));
  inc *= degreesToRadians;

  glBegin( GL_LINE_STRIP );
  for( cur_rot = getMinAngle(); cur_rot < getMaxAngle(); cur_rot += inc )
    glVertex2d( cos( cur_rot) * rad, sin( cur_rot ) * rad );
  // connect the end of the arc
  glVertex2d( cos(getMaxAngle()) * rad, sin(getMaxAngle()) * rad );
  glEnd();
  endGLLineDrawing();
  glPopMatrix();
}

void drawableArc::drawPS( const Graphics &out, psFile &ps,
  const Bbox &bounds, const double &scale ) const
{
  if( lwidth <= 0.0 )
    return;
  ps.newpath();
  setPSColor( ps );
  setupPSLineDrawing( ps );
  Point p = ps.mapPoint( out, bounds, scale, center() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y() << ' '
    << ps.mapRange( scale, ::sqrt( squared_radius() ) ) << ' '
    << getMinAngle() * radiansToDegrees << ' '
    << getMaxAngle() * radiansToDegrees << " arc" << std::endl;
  ps << "stroke" << std::endl;
  endPSLineDrawing( ps );
}

void drawableWedge::drawGL( const Graphics &out ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;
  double rad = (double)::sqrt( squared_radius() );
  double cur_rot;

  // see drawableCircle::drawGL for explanation
  double inc = 360 /
    (4 + 356 * ((out.getWidth() > out.getHeight()) ?
		rad / out.getWidth() : rad / out.getHeight()));
  inc *= degreesToRadians;

  glPushMatrix();
  glTranslated( center().x(), center().y(), 0.0 );

  if( objectFilled ) {
    setGLColor();
    glBegin( GL_TRIANGLE_FAN );
    glVertex2d( 0.0, 0.0 );
    for( cur_rot = getMinAngle(); cur_rot < getMaxAngle(); cur_rot += inc )
      glVertex2d( cos( cur_rot ) * rad, sin( cur_rot ) * rad );
    // connect the end of the arc
    glVertex2d( cos(getMaxAngle()) * rad, sin(getMaxAngle()) * rad );
    glEnd();
  }
  if( lwidth > 0.0 ) { // draw the border
    setGLBorderColor();
    setupGLLineDrawing();
    glBegin( GL_LINE_LOOP );
    glVertex2d( 0.0, 0.0 );
    for( cur_rot = getMinAngle(); cur_rot < getMaxAngle(); cur_rot += inc )
      glVertex2d( cos( cur_rot ) * rad, sin( cur_rot ) * rad );
    // connect the end of the arc
    glVertex2d( cos(getMaxAngle()) * rad, sin(getMaxAngle()) * rad );
    glEnd();
    endGLLineDrawing();
  }
  glPopMatrix();
}

void drawableWedge::drawPS( const Graphics &out, psFile &ps,
  const Bbox &bounds, const double &scale ) const
{
  if( !objectFilled && lwidth <= 0.0 )
    return;
  ps.newpath();
  Point p = ps.mapPoint( out, bounds, scale, center() );
  ps << std::setprecision(4) << p.x() << ' ' << p.y() << ' '
    << ps.mapRange( scale, ::sqrt( squared_radius() ) ) << ' '
    << getMinAngle() * radiansToDegrees << ' '
    << getMaxAngle() * radiansToDegrees << " arc" << std::endl;
  if( objectFilled ) {
    ps << "gsave" << std::endl;
    setPSColor( ps );
    ps << "fill grestore" << std::endl;
  }
  if( lwidth > 0.0 ) {
    setPSBorderColor( ps );
    setupPSLineDrawing( ps );
    ps << "stroke" << std::endl;
    endPSLineDrawing( ps );
  }
}

// drawableImage methods

drawableImage::drawableImage( const Bbox &worldLocation, unsigned int w,
  unsigned int h, bytespp bytes )
  : worldLoc(worldLocation), width(w), height(h), bpp(bytes)
{
  unsigned int size = bpp * width * height;
  buffer = new unsigned char[size];
  memset( buffer, 0, size );
}

drawableImage::drawableImage( const drawableImage &i )
  : worldLoc(i.worldLoc), width(i.width), height(i.height), bpp(i.bpp)
{
  // copy the buffer
  unsigned int size = bpp * width * height;
  buffer = new unsigned char[size];
  memcpy( buffer, i.buffer, size );
}

drawableImage::~drawableImage()
{
  destroy();
}

void drawableImage::destroy()
{
  delete [] buffer;
  buffer = 0;
}

void drawableImage::setPixel( unsigned int x, unsigned int y,
  const unsigned char *col )
{
  if( x >= width || y >= height )
    return;
  if( bpp == BPP1 )
    buffer[y * width + x] = *col;
  else {
    unsigned int p = 3 * (y * width + x);
    buffer[p] = *col;
    buffer[p + 1] = *(col + 1);
    buffer[p + 2] = *(col + 2);
  }
}

const unsigned char * const drawableImage::getPixel( unsigned int x,
  unsigned int y ) const
{
  if( x >= width || y >= height )
    return 0;
  return &buffer[bpp * (y * width + x)];
}

bool drawableImage::fail( FILE *stream )
{
  fclose( stream );
  destroy();
  return false;
}

inline void skipWhiteComment( FILE *in )
{
  static char c;
  while( (c = fgetc( in )) != EOF ) {
    if( c == '#' ) // skip to end of line
      fscanf( in, "%*[^\n]" );
    else if( !isspace( c ) ) {
      ungetc( c, in );
      break;
    }
  }
}

bool drawableImage::loadPPM( const char *filename )
{
  destroy();
  FILE *in = fopen( filename, "rb" );
  if( !in )
    return false;
  // read magic number
  unsigned char pchar;
  if( !fread( &pchar, 1, 1, in ) || pchar != 'P' )
    return fail( in );
  unsigned char magic;
  bool ascii;
  if( !fread( &magic, 1, 1, in ) )
    return fail( in );
  if( magic == '3' )
    ascii = true;
  else if( magic == '6' )
    ascii = false;
  else
    return fail( in );
  skipWhiteComment( in );
  bpp = BPP3;
  if( !fscanf( in, "%d", &width ) || width == 0 )
    return fail( in );
  skipWhiteComment( in );
  if( !fscanf( in, "%d", &height ) || height == 0 )
    return fail( in );
  skipWhiteComment( in );
  unsigned int maxval;
  if( !fscanf( in, "%d", &maxval ) || maxval == 0 )
    return fail( in );

  buffer = new unsigned char[3 * width * height];

  if( ascii ) {
    if( !readAscii( in, maxval ) )
      return fail( in );
  } else if( !readBinary( in ) )
      return fail( in );

  fclose( in );
  return true;
}

bool drawableImage::readAscii( FILE *in, unsigned int maxval )
{
  skipWhiteComment( in );
  unsigned int cnt = 0, size = width * height;
  unsigned int pix[3], sp;
  unsigned int hm1 = height - 1;
  while( !feof( in ) && cnt < size ) {
    if( !fscanf( in, "%d", &pix[0] ) )
      return false;
    skipWhiteComment( in );
    if( !fscanf( in, "%d", &pix[1] ) )
      return false;
    skipWhiteComment( in );
    if( !fscanf( in, "%d", &pix[2] ) )
      return false;
    skipWhiteComment( in );
    if( maxval > 255 ) {
      pix[0] = (unsigned int)(float(pix[0]) / float(maxval)) * 255;
      pix[1] = (unsigned int)(float(pix[1]) / float(maxval)) * 255;
      pix[2] = (unsigned int)(float(pix[2]) / float(maxval)) * 255;
    }
    // because ppm is stored top->bottom and we want bottom->top,
    //  we need to reverse the y-axis
    sp = 3 * ((hm1 - cnt / width) * width + (cnt % width));
    buffer[sp] = (unsigned char)pix[0];
    buffer[sp + 1] = (unsigned char)pix[1];
    buffer[sp + 2] = (unsigned char)pix[2];
    ++cnt;
  }
  if( cnt < size )
    return false;
  return true;
}

bool drawableImage::readBinary( FILE *in )
{
  char c;
  if( !fread( &c, 1, 1, in ) )
    return false;
  if( !isspace( c ) )
    ungetc( c, in );
  unsigned int cnt = 0, size = width * height, sp = 0;;
  unsigned char pix[3];
  unsigned int hm1 = height - 1;
  while( !feof( in ) && cnt < size ) {
    if( fread( pix, 1, 3, in ) < 3 )
      return false;
    sp = 3 * ((hm1 - cnt / width) * width + (cnt % width));
    buffer[sp] = pix[0];
    buffer[sp + 1] = pix[1];
    buffer[sp + 2] = pix[2];
    ++cnt;
  }
  if( cnt < size )
    return false;
  return true;
}

bool drawableImage::savePPM( const char *filename, bool binaryFile ) const
{
  FILE *out = fopen( filename, "wb" );
  if( !out )
    return false;
  fprintf( out, (binaryFile ? "P6" : "P3") );
  fprintf( out, "\n%d\n%d\n255\n", width, height );
  const unsigned char *pix;
  int y, x, cnt = 0;
  // this will flip the y-axis
  for( y = height - 1; y >= 0; --y ) {
    for( x = 0; x < int(width); ++x ) {
      pix = getPixel( x, y );
      if( binaryFile ) {
        if( bpp == BPP1 ) {
          fwrite( pix, 1, 1, out );
          fwrite( pix, 1, 1, out );
          fwrite( pix, 1, 1, out );
        } else
          fwrite( pix, 3, 1, out );
      } else {
        if( bpp == BPP1 ) {
          fprintf( out, "%d %d %d ", *pix, *pix, *pix );
        } else
          fprintf( out, "%d %d %d ", pix[0], pix[1], pix[2] );
        if( !(cnt % 6) )
          fprintf( out, "\n" );
        ++cnt;
      }
    }
  }
  fclose( out );
  return true;
}

// FIXME: investigate the alternative of drawing as a texture,
//  which would probably be a much faster operation; the only
//  problem is that the image must then be 2^n x 2^n, so we
//  would have to pad it and use texture coordinates to draw
//  only the relevant stuff.
void drawableImage::drawGL( const Graphics &out ) const
{
  // find scaling factors
  float ppu;
  out.pixelsPerUnit( ppu );
  float pxf = ppu * float(worldLoc.xmax() - worldLoc.xmin()) / float(width);
  float pyf = ppu * float(worldLoc.ymax() - worldLoc.ymin()) / float(height);
  setGLColor();
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
  glPixelZoom( pxf, pyf );
  
  // trick so that when the raster pos is offscreen the bitmap still draws
  // with help from http://www.mesa3d.org/brianp/TR.html
  GLdouble mview[16], proj[16], x, y, z;
  GLint vport[4];
  glGetDoublev( GL_MODELVIEW_MATRIX, mview );
  glGetDoublev( GL_PROJECTION_MATRIX, proj );
  vport[0] = vport[1] = 0;
  vport[2] = out.getPixWidth();
  vport[3] = out.getPixHeight();
  if( gluProject( worldLoc.xmin(), worldLoc.ymin(), 0.0, mview, proj,
    vport, &x, &y, &z ) )
  {
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    glOrtho( 0.0, out.getPixWidth(), 0.0, out.getPixHeight(), 0.0, 1.0 );
    glRasterPos3d( 0.0, 0.0, -z );
    unsigned char tbm[1] = {0};
    glBitmap( 1, 1, 0.0, 0.0, x, y, tbm );
    glPopMatrix();
    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
  }

  // now draw the pixels
  glDrawPixels( width, height, (bpp == BPP3 ? GL_RGB : GL_LUMINANCE),
    GL_UNSIGNED_BYTE, buffer );
}


void drawableImage::drawPS( const Graphics &out, psFile &ps,
  const Bbox &bounds, const double &scale ) const
{
  ps << "/picstr " << width << " string def" << std::endl
    << "gsave" << std::endl;
  Point p, p2;
  p = ps.mapPoint( out, bounds, scale, Point(worldLoc.xmin(),worldLoc.ymin()) );
  p2 = ps.mapPoint( out, bounds, scale, Point(worldLoc.xmax(),worldLoc.ymax()) );
  ps << std::setprecision(4) << p.x() << ' ' << p.y() << " translate" << std::endl
    << (p2.x() - p.x()) << ' ' << (p2.y() - p.y()) << " scale" << std::endl
    << width << ' ' << height << ' ' << 8 << std::endl
    << "[" << width << " 0 0 " << height << " 0 0]" << std::endl
    << "{currentfile picstr readhexstring pop}" << std::endl
    << "false " << bpp << " colorimage" << std::endl;
  ps.setf( std::ios::hex, std::ios::basefield );
  char fill = ps.fill( '0' );
  int size = width * height * bpp;
  for( int i = 0; i < size; ++i ) {
    ps.width( 2 );
    ps << (buffer[i] & 0xff);
    if( !((i + 1) % width) )
      ps << std::endl;
  }
  ps.setf( std::ios::dec, std::ios::basefield );
  ps.fill( fill );
  ps << "grestore" << std::endl;
}
