/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: geometry.cpp,v 1.6 2003/05/27 17:14:23 beevek Exp $
*/

#include <iostream>
#include <dolt-lite.h>

using namespace dolt;

// only implement this functionality if CGAL isn't going
//  to be used
#ifndef USE_CGAL

#include <math.h>
// Point methods

bool Point::operator==( const Point &p ) const
{
  return (ix == p.ix && iy == p.iy);
}

bool Point::operator!=( const Point &p ) const
{
  return !(*this == p);
}

Point Point::rotate(double th) const
{
  return Point(ix*cos(th)-iy*sin(th), ix*sin(th)+iy*cos(th));
}

double Point::length() const
{
  return sqrt(ix*ix + iy*iy); 
}

double dolt::dot(const Point &p, const Point &q)
{
  return p.x()*q.x() + p.y()*q.y();
}

double dolt::cross(const Point &p, const Point &q)
{
  return p.x()*q.y() - p.y()*q.x();
}

Point dolt::unitNormal(double th)
{
  return Point(cos(th), sin(th));
}

std::ostream & dolt::operator<<(std::ostream & out, const dolt::Point &p)
{
  return out << p.x() << ' ' << p.y();
}

std::istream & dolt::operator>>(std::istream & in, dolt::Point &p)
{
  double x, y;
  in >> x >> y;
  p = dolt::Point(x,y);
  return in;
}

// Segment methods

std::ostream & dolt::operator<<(std::ostream &out, const dolt::Segment &s)
{
  return out << s.start() << "  " << s.end();
}

std::istream & dolt::operator>>(std::istream &in, dolt::Segment &s)
{
  double w, x, y, z;
  in >> w >> x >> y >> z;
  s = dolt::Segment(dolt::Point(w,x), dolt::Point(y,z));
  return in;
}

// Polygon methods

Segment Polygon::Edge_const_iterator::operator*()
{
  std::vector<Point>::const_iterator endpt = cur;
  ++endpt;
  if( endpt == container->end() )
    endpt = container->begin();
  return Segment(*cur,*endpt);
}

Polygon::Edge_const_iterator Polygon::Edge_const_iterator::operator++( int )
{
  Edge_const_iterator temp = *this;
  ++*this;
  return temp;
}

Polygon::Edge_const_iterator Polygon::edges_begin() const
{
  return Edge_const_iterator(this,begin());
}

Polygon::Edge_const_iterator Polygon::edges_end() const
{
  return Edge_const_iterator(this,end());
}

Bbox Polygon::bbox() const
{
  std::vector<Point>::const_iterator i = begin();
  double xmin, xmax, ymin, ymax;
  xmin = xmax = i->x();
  ymin = ymax = i->y();
  for( ++i; i != end(); ++i ) {
    if( i->x() < xmin )
      xmin = i->x();
    if( i->x() > xmax )
      xmax = i->x();
    if( i->y() < ymin )
      ymin = i->y();
    if( i->y() > ymax )
      ymax = i->y();
  }
  return Bbox(xmin, ymin, xmax, ymax);
}

std::ostream & dolt::operator<<(std::ostream & out, const Polygon &p)
{
  out << p.size() << ' ';
  Polygon::Edge_const_iterator i = p.edges_begin();
  for( ; i != p.edges_end(); ++i )
    out << i->start() << ' ';
  return out;
}

std::istream& dolt::operator>>(std::istream& in, Polygon& p)
{
  int n;
  
  in >> n; 			// get number of points
  for (int k = 0; k < n; k++) {
    Point q;
    in >> q;
    p.push_back(q);
  }
  return in;
}

#endif // USE_CGAL

Bbox dolt::stretchBbox( const Bbox &box, float scaleX, float scaleY )
{
  double dx = box.xmax() - box.xmin();
  double dy = box.ymax() - box.ymin();
  Point c(dx / 2 + box.xmin(), dy / 2 + box.ymin());
  dx *= 0.5 * scaleX;
  dy *= 0.5 * scaleY;
  return Bbox(c.x() - dx, c.y() - dy, c.x() + dx, c.y() + dy);
}













