/*
  This file is part of dolt.

  dolt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  dolt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with dolt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  $Id: graphics.cpp,v 1.9 2004/01/27 02:38:32 whuang Exp $
*/

#include <GL/glut.h>
#include <algorithm> // for find
#include <dolt-lite.h>
#include <iostream> // for debugging
using namespace dolt;

// must initialize these here
Graphics * Graphics::currentInstance = 0;
bool Graphics::glutInitialized = false;

// constructors and destructors; note that only one instance
//  of this class may be instantiated

Graphics::Graphics()
  : worldbox(0.0,0.0,10.0,10.0),
    zoomPercent(1.0f), panPercentX(0.0), panPercentY(0.0),
    title(0), userKeyboard(0), userMouse(0), userMotion(0),
    userIdle(0), animateCallback(0)
{
  if( !currentInstance )
    currentInstance = this;
}

Graphics::Graphics( const Bbox &world )
  : worldbox(world),
    zoomPercent(1.0f), panPercentX(0.0), panPercentY(0.0),
    title(0), userKeyboard(0), userMouse(0), userMotion(0),
  userIdle(0)
{
  if( !currentInstance )
    currentInstance = this;
}

Graphics::~Graphics()
{
  if( currentInstance == this )
    currentInstance = 0;
  if( title )
    delete [] title;
}

// initialize glut
void Graphics::initGL( int *argc, char **argv, const char *name,
  int winWidth, int winHeight, int winPosX, int winPosY )
{
  if( winHeight <= 0 ) {
    // figure out height based on width and current
    //  world bounding box
    double wx = worldbox.xmax() - worldbox.xmin();
    double wy = worldbox.ymax() - worldbox.ymax();
    winHeight = int( ((double)winWidth / wx) * wy );
  }
  pixW = winWidth;
  pixH = winHeight;

  title = (char *)strdup( name ? name : argv[0] );

  if( glutInitialized )
    return;

  glutInit( argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA );
  glutInitWindowSize( pixW, pixH );
  glutInitWindowPosition( winPosX, winPosY );
  glutCreateWindow( title );
  glutReshapeFunc( reshape );
  glutDisplayFunc( display );
  glutKeyboardFunc( keyboard );
  glutSpecialFunc( specialKeyboard );
  glMatrixMode( GL_PROJECTION );
  glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
  glEnable( GL_POINT_SMOOTH );
  glEnable( GL_LINE_SMOOTH );

  glEnable( GL_BLEND ); // FIXME: this probably shouldn't be global

  glutInitialized = true;

  setupWindow();
}

void Graphics::setupWindow()
{
  glClearColor( bgColor.r(), bgColor.g(), bgColor.b(), bgColor.a() );
  glLoadIdentity();
  gluOrtho2D( worldbox.xmin(), worldbox.xmax(),
    worldbox.ymin(), worldbox.ymax() );
  glutReshapeWindow( pixW, pixH );
  glutSetWindowTitle( title ? title : "" );
  glutIdleFunc( userIdle ? userIdle : 0 );
  glutMouseFunc( userMouse ? userMouse : 0 );
  glutPassiveMotionFunc( userMotion ? userMotion : 0 );
}

void Graphics::setCurrentScene( Graphics &g )
{
  currentInstance = &g;
  g.setupWindow();
}

void Graphics::setBackgroundColor( const Color &c )
{
  bgColor = c;
  glClearColor( bgColor.r(), bgColor.g(), bgColor.b(), bgColor.a() );
}

void Graphics::runEventLoop()
{
  glutMainLoop();
}

// in-class default event handlers for display,
//  reshape, keyboard, mouse and motion events

void Graphics::reshape( int w, int h )
{
  if( !currentInstance )
    return;
  currentInstance->pixH = h;
  currentInstance->pixW = w;

  glViewport( 0, 0, w, h );
  currentInstance->zoom( currentInstance->zoomPercent );
}

void Graphics::display()
{
  if( !currentInstance )
    return;
  if( currentInstance->animateCallback )
    currentInstance->animateCallback();
  glClear( GL_COLOR_BUFFER_BIT );
  std::list<const objectList *>::const_iterator i;
  objectList::const_iterator j;

  for( i = currentInstance->objectLists.begin();
       i != currentInstance->objectLists.end(); i++ ) {
    if ((*i)->visible()) 
      for( j = (*i)->begin(); j != (*i)->end(); j++ ) 
	(*j)->drawGL( *currentInstance );
  }
  glutSwapBuffers();
  if( currentInstance->animateCallback )
    glutPostRedisplay();
}

void Graphics::keyboard( unsigned char key, int x, int y )
{
  if( !currentInstance )
    return;
  if( !currentInstance->userKeyboard ) {
    if( key == 'q' )
      exit( 0 );
    else if( key == 'h' ) {
      std::cout << "Dolt Default Keybindings:" << std::endl
        << " q: Quit" << std::endl
        << " h: Help" << std::endl;
      currentInstance->printControlKeyBindings();
    }
  } else {
    currentInstance->userKeyboard( key );
    glutPostRedisplay(); // assume a redraw is needed
  }
}

void Graphics::specialKeyboard( int key, int x, int y )
{
  switch( key ) {
  case GLUT_KEY_LEFT:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->pan( 0.25, 0.0 );
    else
      currentInstance->pan( 0.05, 0.0 );
    break;
  case GLUT_KEY_UP:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->pan( 0.0, -0.25 );
    else
      currentInstance->pan( 0.0, -0.05 );
    break;
  case GLUT_KEY_RIGHT:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->pan( -0.25, 0.0 );
    else
      currentInstance->pan( -0.05, 0.0 );
    break;
  case GLUT_KEY_DOWN:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->pan( 0.0, 0.25 );
    else
      currentInstance->pan( 0.0, 0.05 );
    break;
  case GLUT_KEY_PAGE_UP:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->zoom( currentInstance->zoomPercent * 2.0 );
    else
      currentInstance->zoom( currentInstance->zoomPercent * 1.10 );
    break;
  case GLUT_KEY_PAGE_DOWN:
    if( glutGetModifiers() & GLUT_ACTIVE_SHIFT )
      currentInstance->zoom( currentInstance->zoomPercent * 0.5 );
    else
      currentInstance->zoom( currentInstance->zoomPercent * 0.90 );
    break;
  case GLUT_KEY_HOME:
    currentInstance->zoom( 1.0 );
    currentInstance->pan( -currentInstance->panPercentX,
      -currentInstance->panPercentY );
    break;
  }
  glutPostRedisplay();
}

void Graphics::setKeyboardHandler( void (*kb)( unsigned char key ) )
{
  userKeyboard = kb;
}

void Graphics::setMouseHandler( void (*mouse)( int button, int state, int x, int y ) )
{
  userMouse = mouse;
  glutMouseFunc( mouse );
}

void Graphics::setMotionHandler( void (*motion)( int x, int y ) )
{
  userMotion = motion;
  glutPassiveMotionFunc( motion );
}

void Graphics::setIdleFunction( void (*idle)( void ) )
{
  userIdle = idle;
  if( !animateCallback )
    glutIdleFunc( idle );
}

void Graphics::setAnimateCallback( void (*processing)( void ) )
{
  animateCallback = processing;
  if( processing )
    glutIdleFunc( display );
  else
    glutIdleFunc( userIdle );
}

void Graphics::printControlKeyBindings() const
{
  std::cout << " Page Up: Zoom in 10% (with Shift, zoom in 50%)" << std::endl
    << " Page Down: Zoom out 10% (with Shift, zoom out 50%)" << std::endl
    << " Arrow Keys: Pan view by 5% of window (with Shift, 50% of window)" << std::endl
    << " Home: Return to original view (0% zoom, 0% pan)" << std::endl;
}

void Graphics::zoom( float newZoom )
{
  double dx = worldbox.xmax() - worldbox.xmin();
  double dy = worldbox.ymax() - worldbox.ymin();
  double xcenter = worldbox.xmin() + (dx / 2.0 - dx * panPercentX);
  double ycenter = worldbox.ymin() + (dy / 2.0 - dy * panPercentY);
  double dxz = dx / newZoom;
  double dyz = dy / newZoom;
  double dxzhalf = dxz / 2.0;
  double dyzhalf = dyz / 2.0;

  double waspect = double(pixW) / dxz;
  double haspect = double(pixH) / dyz;
  double half;

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  if( haspect < waspect ) {
    half = (dxz * (waspect / haspect) - dxz) / 2.0;
    gluOrtho2D( -half + xcenter - dxzhalf, dxzhalf + half + xcenter,
      ycenter - dyzhalf, ycenter + dyzhalf );
  } else {
    half = (dyz * (haspect / waspect) - dyz) / 2.0;
    gluOrtho2D( xcenter - dxzhalf, xcenter + dxzhalf,
      -half + ycenter - dyzhalf, dyzhalf + half + ycenter );
  }

  // original: don't keep aspect ratio the same
  //gluOrtho2D( xcenter - dxz, xcenter + dxz,
  //  ycenter - dyz, ycenter + dyz );

  zoomPercent = newZoom;

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  glutPostRedisplay();
}

void Graphics::pan( float panX, float panY )
{
  panPercentX += panX;
  panPercentY += panY;
  zoom( zoomPercent );
}

void Graphics::pixelsPerUnit( float &ppu  ) const
{
  double dx = worldbox.xmax() - worldbox.xmin();
  double dy = worldbox.ymax() - worldbox.ymin();
  if( double(pixW) / dx > double(pixH) / dy )
    ppu = float(pixH) * zoomPercent / float(worldbox.ymax() - worldbox.ymin());
  else
    ppu = float(pixW) * zoomPercent / float(worldbox.xmax() - worldbox.xmin());
}

// add a list of objects to be drawn
void Graphics::addObjectList( const objectList &olist )
{
  objectLists.push_back( &olist );
}

// delete a list of objects from the set of things to
//  be drawn
void Graphics::delObjectList( const objectList &olist )
{
  // since our list of lists is represented by
  //  a linked list, the only real way to do this
  //  is to iterate through everything until we
  //  find the right list to delete
  std::list<const objectList *>::iterator i;
  i = std::find( objectLists.begin(), objectLists.end(), &olist );
  if( i != objectLists.end() )
    objectLists.erase( i );
}

Point Graphics::mapPoint( const Bbox &bounds, const double &scale,
  const Point &p ) const
{
  // must perform these calculations to take into
  //  account zooming and panning
  double dx = worldbox.xmax() - worldbox.xmin();
  double dy = worldbox.ymax() - worldbox.ymin();
  double minX = (dx / 2.0 - dx * panPercentX) - (dx / zoomPercent / 2.0);
  double minY = (dy / 2.0 - dy * panPercentY) - (dy / zoomPercent / 2.0);

  return Point( ((p.x() - minX) * scale + bounds.xmin()),
    ((p.y() - minY) * scale + bounds.ymin()) );
}

// utility for writePS; find "scaling factor" between worldBounds
// and newBounds
double findScale( const Bbox &worldBounds, const Bbox &newBounds )
{
  double wx = worldBounds.xmax() - worldBounds.xmin();
  double wy = worldBounds.ymax() - worldBounds.ymin();
  double nx = newBounds.xmax() - newBounds.xmin();
  double ny = newBounds.ymax() - newBounds.ymin();
  if( nx / wx > ny / wy )
    return ny / wy;
  else
    return nx / wx;
}

void Graphics::writePS( psFile &out, const Bbox &bounds ) const
{
  if( !out.is_open() )
    return;
  double scale = findScale( worldbox, bounds );
  std::list<const objectList *>::const_iterator i;
  objectList::const_iterator j;
  for( i = objectLists.begin(); i != objectLists.end(); i++ )
    for( j = (*i)->begin(); j != (*i)->end(); j++ )
      (*j)->drawPS( *this, out, bounds, scale );
}

bool Graphics::writePPM( const char *filename ) const
{
  drawableImage img(Bbox(0.0,0.0,0.0,0.0), pixW, pixH, drawableImage::BPP3);
  glPixelStorei( GL_PACK_ALIGNMENT, 1 );
  glReadPixels( 0, 0, pixW, pixH, GL_RGB, GL_UNSIGNED_BYTE, img.getBuffer() );
  return img.savePPM( filename, true );
}









